#ifndef _TS_Buffer_H
#define _TS_Buffer_H

#include "Prefix.h"
#include "EARTH_PT.h"
#include "TS_Stream.h"
#include "TS_Utility.h"
#include "OS_Memory.h"
#include "OS_Thread.h"

namespace EARTH {
namespace TS {
	class Buffer {
	public:
		Buffer(PT::Device *device) : mDevice(device)
		{
			_ASSERTE(((OS::Memory::PAGE_SIZE * PT::Device::BUFFER_PAGE_COUNT) % BLOCK_COUNT) == 0);

			mMemory = NULL;
			for (uint i=0; i<sizeof(mStream)/sizeof(*mStream); i++) {
				mStream[i] = NULL;
			}
			mThread = NULL;
			mQuit = false;
		}

		~Buffer()
		{
			Stop();
		}

		void Run()
		{
			status status;

			if (mMemory == NULL) {
				// e|̈mۂ
				uint pageCount = (BLOCK_SIZE + OS::Memory::PAGE_SIZE-1) / OS::Memory::PAGE_SIZE;
				mMemory = new OS::Memory(pageCount);
			}

			// Xg[𐶐
			for (uint i=0; i<sizeof(mStream)/sizeof(*mStream); i++) {
				if (mStream[i] == NULL) {
					mStream[i] = new Stream(i/2, (i%2 == 0) ? L'S' : L'T');
				}
			}

			// obt@mۂĂȂꍇ͊mۂ
			PT::Device::BufferInfo bufferInfo;
			status = mDevice->GetBufferInfo(&bufferInfo);
			if (Error(L"Device::GetBufferInfo()", status)) return;

			if (bufferInfo.VirtualSize == 0) {
				bufferInfo.VirtualSize  = VIRTUAL_SIZE;
				bufferInfo.VirtualCount = VIRTUAL_COUNT;
				bufferInfo.LockSize     = LOCK_SIZE;
				status = mDevice->SetBufferInfo(&bufferInfo);
				if (Error(L"Device::SetBufferInfo()", status)) return;
			}

			// DMA ]ǂ܂Ői񂾂𒲂ׂ邽߁AeubN̖ 0 ŃNA
			for (uint i=0; i<VIRTUAL_COUNT; i++) {
				for (uint j=0; j<VIRTUAL_SIZE; j++) {
					for (uint k=0; k<BLOCK_COUNT; k++) {
						Clear(i, j, k);
					}
				}
			}

			// ]JE^Zbg
			status = mDevice->ResetTransferCounter();
			if (Error(L"Device::ResetTransferCounter()", status)) return;

			// ]JE^CNg
			for (uint i=0; i<VIRTUAL_SIZE*VIRTUAL_COUNT; i++) {
				status = mDevice->IncrementTransferCounter();
				if (Error(L"Device::IncrementTransferCounter()", status)) return;
			}

			// DMA ]
			status = mDevice->SetTransferEnable(true);
			if (Error(L"Device::SetTransferEnable()", status)) return;

			// XbhJn
			mQuit = false;
			_ASSERTE(mThread == NULL);
			mThread = new OS::Thread(Thread_, this);
		}

		void Stop()
		{
			status status;

			// DMA ]Ăꍇ͒~
			bool transferEnabled;
			status = mDevice->GetTransferEnable(&transferEnabled);
			if (Error(L"Device::GetTransferEnable()", status)) return;

			if (transferEnabled) {
				status = mDevice->SetTransferEnable(false);
				Error(L"Device::SetTransferEnable()", status);
			}

			// Xbh~
			if (mThread) {
				mQuit = true;
				delete_(mThread);
			}

			// Xg[
			for (uint i=0; i<sizeof(mStream)/sizeof(*mStream); i++) {
				delete_(mStream[i]);
			}

			// e|̈
			delete_(mMemory);

			// DMA obt@mۂĂꍇ͉
			PT::Device::BufferInfo bufferInfo;
			status = mDevice->GetBufferInfo(&bufferInfo);
			Error(L"Device::GetBufferInfo()", status);

			if (bufferInfo.VirtualSize) {
				status = mDevice->SetBufferInfo(NULL);
				Error(L"Device::SetBufferInfo(NULL)", status);
			}
		}

	protected:
		enum {
			VIRTUAL_SIZE  =  4,
			VIRTUAL_COUNT = 16,
			LOCK_SIZE     =  4
		};

		enum {
			BLOCK_COUNT = PT::Device::BUFFER_PAGE_COUNT/*2*/,
			BLOCK_SIZE  = OS::Memory::PAGE_SIZE * PT::Device::BUFFER_PAGE_COUNT / BLOCK_COUNT
		};
		// DMA ̃o[Xg 4096 oCgłB̂߁ABLOCK_COUNT  PT::Device::BUFFER_PAGE_COUNT 傫Ă
		// CeVPʂ͂قƂǓ܂B
		// ܂ÃTvł (OS::Memory::PAGE_SIZE * PT::Device::BUFFER_PAGE_COUNT) % BLOCK_COUNT  0 ł邱ƂKvłB

		// (obt@̍\)
		//                    +--------------+------------+------------+
		//                    | VirtualIndex | ImageIndex | BlockIndex |
		//                    +--------------+------------+------------+-------------------+
		// GetBufferPtr(0) -> |            0 |          0 |          0 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		//                    |            0 |          0 |          1 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		//                    |            0 |          1 |          0 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		//                    |            0 |          1 |          1 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		//                    |            0 |          2 |          0 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		//                    |            0 |          2 |          1 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		//                    |            0 |          3 |          0 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		//                    |            0 |          3 |          1 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		// GetBufferPtr(1) -> |            1 |          0 |          0 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		//                    :              :            :            :                   :
		//                    :              :            :            :                   :
		//                    :              :            :            :                   :
		//                    +--------------+------------+------------+-------------------+
		//                    |           15 |          3 |          1 | BLOCK_SIZE oCg |
		//                    +--------------+------------+------------+-------------------+
		//
		// (l͈̔)
		// 0  VirtualIndex  VIRTUAL_COUNT
		// 0  ImageIndex    VIRTUAL_SIZE
		// 0  BlockIndex    BLOCK_COUNT
		// 
		// (]JE^)
		// ]JE^̑ OS::Memory::PAGE_SIZE * PT::Device::BUFFER_PAGE_COUNT oCgƂɍs܂B

		PT::Device *const mDevice;
		OS::Memory *mMemory;
		uint mVirtualIndex;
		uint mImageIndex;
		uint mBlockIndex;
		Stream *mStream[2*2];
		OS::Thread *mThread;
		volatile bool mQuit;

		// 1ubN DMA ]I邩 mQuit  true ɂȂ܂ő҂
		bool WaitBlock()
		{
			bool b = true;

			while (true) {
				if (mQuit) {
					b = false;
					break;
				}

				// ubN̖ 0 łȂ΁ÃubN DMA ]ƂɂȂ
				if (Read(mVirtualIndex, mImageIndex, mBlockIndex)) break;

				OS::Thread::Sleep((1)*1000*1000);
			}
			//::wprintf(L"(mVirtualIndex, mImageIndex, mBlockIndex) = (%d, %d, %d) ̓]I܂B\n", mVirtualIndex, mImageIndex, mBlockIndex);

			return b;
		}

		// 1ubÑf[^e|̈ɃRs[BCPU 猩 DMA obt@̓LbVȂ߁A
		// LbV̈ɃRs[ĂANZXƌ܂܂B
		void CopyBlock()
		{
			status status;

			void *voidPtr;
			status = mDevice->GetBufferPtr(mVirtualIndex, &voidPtr);
			if (Error(L"Device::GetBufferPtr()", status)) return;

			uint *srcPtr = static_cast<uint *>(voidPtr);
			_ASSERTE(srcPtr);

			srcPtr += Offset(mImageIndex, mBlockIndex, 0);

			void *dstPtr = mMemory->Ptr();
			_ASSERTE(dstPtr);

			::memcpy(dstPtr, srcPtr, BLOCK_SIZE);

			// Rs[ÎŁAubN̖ 0 ɂ܂B
			srcPtr[BLOCK_SIZE/sizeof(*srcPtr)-1] = 0;

			if (BLOCK_COUNT <= ++mBlockIndex) {
				mBlockIndex = 0;

				// ]JE^ OS::Memory::PAGE_SIZE * PT::Device::BUFFER_PAGE_COUNT oCgƂɃCNg܂B
				status = mDevice->IncrementTransferCounter();
				if (Error(L"Device::IncrementTransferCounter()", status)) return;

				if (VIRTUAL_SIZE <= ++mImageIndex) {
					mImageIndex = 0;
					if (VIRTUAL_COUNT <= ++mVirtualIndex) {
						mVirtualIndex = 0;
					}
				}
			}
		}

		void Clear(uint virtualIndex, uint imageIndex, uint blockIndex)
		{
			void *voidPtr;
			status status = mDevice->GetBufferPtr(virtualIndex, &voidPtr);
			if (Error(L"Device::GetBufferPtr()", status)) return;

			uint *ptr = static_cast<uint *>(voidPtr);
			_ASSERTE(ptr);
			ptr[Offset(imageIndex, blockIndex)] = 0;
		}

		uint Read(uint virtualIndex, uint imageIndex, uint blockIndex) const
		{
			void *voidPtr;
			status status = mDevice->GetBufferPtr(virtualIndex, &voidPtr);
			if (Error(L"Device::GetBufferPtr()", status)) return 0;

			volatile const uint *ptr = static_cast<volatile const uint *>(voidPtr);
			_ASSERTE(ptr);
			return ptr[Offset(imageIndex, blockIndex)];
		}

		uint Offset(uint imageIndex, uint blockIndex, uint additionalOffset = BLOCK_SIZE-sizeof(uint)) const
		{
			blockIndex += BLOCK_COUNT * imageIndex;
			uint offset = (BLOCK_SIZE * blockIndex + additionalOffset) / sizeof(uint);

			return offset;
		}

		// ubÑ}CNpPbgeXg[ɕz
		bool DispatchBlock()
		{
			const uint *ptr = static_cast<uint *>(mMemory->Ptr());
			_ASSERTE(ptr);

			for (uint i=0; i<BLOCK_SIZE/sizeof(*ptr); i++) {
				uint packet = ptr[i];
				_ASSERTE(packet);
				uint packetId    = BIT_SHIFT_MASK(packet, 29, 3);
				uint packetError = BIT_SHIFT_MASK(packet, 24, 1);

				if (packetError) {
					// G[̌𒲂ׂ
					PT::Device::TransferInfo info;
					status status = mDevice->GetTransferInfo(&info);
					if (Error(L"Device::GetTransferInfo()", status)) return false;

					if (info.TransferCounter1) {
						::wprintf(L"]JE^ 1 ȉɂȂ܂B\n");
					} else if (info.BufferOverflow) {
						::wprintf(L"PCI oX𒷊ɓnmۂłȂ߁A{[h FIFO ܂B\n");
					} else {
						::wprintf(L"]G[܂B\n");
					}
					::wprintf(L"t@Cւ̏o𒆎~܂B\n");
					return false;
				}

				if ((1 <= packetId) && (packetId <= 4)) {
//				if ((4 <= packetId) && (packetId <= 4)) {
					mStream[packetId-1]->MicroPacket(packet);

					/*
					static uint sCounter;
					static bool sSleep = true;
					if ((sCounter % (1024*8)) == 0) {
						mDevice->SetTunerSleep(0, PT::Device::ISDB_T, sSleep);
						sSleep = !sSleep;
					}
					sCounter++;
					*/
				}
			}

			return true;
		}

		static void Thread_(void *argument)
		{
			Buffer *that = static_cast<Buffer *>(argument);
			that->Thread();
		}

		void Thread()
		{
			mVirtualIndex = 0;
			mImageIndex = 0;
			mBlockIndex = 0;

			while (true) {
				bool b;
				
				b = WaitBlock();
				if (b == false) break;

				CopyBlock();

				b = DispatchBlock();
				if (b == false) break;
			}
		}
	};
}
}

#endif

